#include "types.h"
#include "riscv.h"
#include "defs.h"
#include "param.h"
#include "memlayout.h"
#include "spinlock.h"
#include "proc.h"
#include "timer.h"
#include "sysinfo.h"
#include "signal.h"

uint64
sys_exit(void)
{
  int n;
  argint(0, &n);
  exit(n);
  return 0;  // not reached
}

uint64
sys_getpid(void)
{
  return myproc()->pid;
}

uint64
sys_fork(void)
{
  return fork();
}

uint64
sys_wait(void)
{
  uint64 p;
  argaddr(0, &p);
  return wait(p);
}

uint64
sys_sbrk(void)
{
  uint64 addr;
  int n;

  argint(0, &n);

  addr = myproc()->sz;
  if(growproc(n) < 0)
    return -1;
  return addr;
}

uint64
sys_sleep(void)
{
  int n;
  uint ticks0;

  argint(0, &n);
  if(n < 0)
    n = 0;
  acquire(&tickslock);
  ticks0 = ticks;
  while(ticks - ticks0 < n){
    if(killed(myproc())){
      release(&tickslock);
      return -1;
    }
    sleep(&ticks, &tickslock);
  }
  release(&tickslock);
  return 0;
}

uint64
sys_nanosleep(void){
  uint64 addr; // User pointer to the timespec structure
  argaddr(0, &addr);
  if (addr == 0)
      return -1;

  struct timespec ts;
  if (copyin(myproc()->pagetable, (char *)&ts, addr, sizeof(ts)) < 0)
      return -1;

  int total_ticks = (ts.tv_sec * TIMER_DIV) + (ts.tv_nsec / 10000000);
  if (total_ticks <= 0)
      return 0; // No need to sleep if non-positive duration

  acquire(&tickslock);
  uint64 start_ticks = ticks;

  while (ticks - start_ticks < total_ticks)
  {
    if(killed(myproc())){
      release(&tickslock);
      return -1;
    }
    sleep(&ticks, &tickslock);
  }
  release(&tickslock);
  return 0;  
}

uint64
sys_kill(void)
{
  int pid;

  argint(0, &pid);
  return kill(pid);
}

// return how many clock tick interrupts have occurred
// since start.
uint64
sys_uptime(void)
{
  uint xticks;

  acquire(&tickslock);
  xticks = ticks;
  release(&tickslock);
  return xticks;
}

uint64
sys_trace(void){
  uint64 mask;
  
  arguint64(0, &mask);
  struct proc *p;
  p = myproc();
  if(mask == 0){
    return -1;
  }
  p->tracemask = mask;
  return 0;
}

uint64
sys_sysinfo(void){
  uint64 p;
  arguint64(0, &p);
  if(p == 0) 
  return -1;
  struct sysinfo result;
  result.freemem = computeFreeMemory();
  result.nproc = computeProcessNum();
  return copyout(myproc()->pagetable, p, (char*)(&result), sizeof(result));
}

uint64 sys_sigalarm(void){
  int alarm_interval;
  argint(0, &alarm_interval);
  uint64 handler;
  argaddr(1, &handler);
  struct proc *p = myproc();
  p->alarm_int = alarm_interval;
  p->alarmhandler = (void (*)()) handler;
  return 0;
}

uint64 sys_sigreturn(void){
  struct proc *p;
  p = myproc();
  if(p->inAlarm == 1){
    p->inAlarm = 0; //clearing the blocking bit
    p->ticksSinceLastAlarm = 0; // clearing the pending bit
    //since we saved the context of uthread in the user stack, we need to load it back to the trapframe!
    restoreContext(p, (struct u_context_t *)p->signal_ucontext_ptr);
    return 0;
  }else {
    panic("sys_sigreturn panic!\n");
  }
  return -1;
}